/*
** NAME
**     saif_enums.c -- Illustra support for SAIF enumerations
**
** DESCRIPTION
**     Enums are implemented as an mi_integer because people expect enumerated
**     types to be an int.  If the number of values gets anywhere big enough 
**     to fill out an int, the input strategy below will be very inefficient.
**
**     Every enum calls the same SAIF entry point in/out function.
**     If you add a new enum, add an entry to ENUM_LIST below so the
**     help functions will output info.
**
**     Main SAIF functions:
**          SaifEnumHelp_l   List implemented enums.
**          SaifEnumHelp     Outputs valid values for an enum.
**          SaifEnumIn       Basic input routine.
**          SaifEnumOut      Basic output routine.
**
** BUILD INSTRUCTIONS
**     DEC Alpha:
**            cc -I$MI_HOME/h -c saif_enums.c
**            ld -shared -expect_unresolved '*' -o saif_enums.so saif_enums.o
**
**     Solaris:
**            cc -I$MI_HOME/h -c -K PIC saif_enums.c
**            ld -dy -G -o saif_enums.so saif_enums.o
**
**     SUN OS:
**            cc -I$MI_HOME/h -c -PIC saif_enums.c
**            ld -o saif_enums.so saif_enums.o
**
**     chmod a+x saif_enums.so
**
** MODS
**     04/13/94  jta  Removed gt,le,eq etc code.
**                    Added CellShape, IndexScheme, UnitMeaning
**     04/14/94  jta  Cleaned up output routines.
**                    Added SaifEnumHelp_l function.
**     05/23/94  jta  Code reorg--preparing to add rest used by tahoe demo.
**     06/07/94  jta  Upgraded to Illustra 2.0.12, standardized naming 
**                    convention so create type statements can be automatically
**                    generated, put int ops code back in for base types.
**     06/20/94  jta  Added rest of enums in tahoe demo set.
**     10/07/94  jta  Added SatelliteBands.
**     10/20/94  jta  UnitMeaning expanded for DigitalNumber and temperature
**                    units, removed TempUnitMeaning.
**     10/27/94  jta  added saif_Release().
**
** $Header: /users/vis/jta/bigfoot/saif/src/RCS/saif_enums.c,v 1.7 1994/10/27 16:49:02 jta Exp $
*/

#include "saif_enums.h"
#include <stdio.h>
#include <string.h>

#define SAIF_RELEASE "3.2"
#define MAX_ENUM_SIZE 30     /* 'boustrophedonByColumn' is the longest */

static char *rcsdate = "$Date: 1994/10/27 16:49:02 $";

mi_text *
saif_Release()
{
     char       saif_release[100];
     mi_text    *retval;

     sprintf(saif_release, "Release %s (SAIF %s)", rcsdate, SAIF_RELEASE);

     retval = mi_string_to_text(saif_release);
     return (retval);
}

char *CellShape_enum[] =                  /* <<<<<<<<<< CellShape >>>>>>>>>> */
{ "point", "rectangle", "ellipse", "hexagon", "triangle", };
#define CellShapeN (sizeof (CellShape_enum) / sizeof (char *))

SaifEnum
CellShapeIn(s, fparam)
mi_text   *s;
MI_FPARAM *fparam;
{ return(SaifEnumIn("CellShape", CellShape_enum, CellShapeN, s, fparam)); }

mi_text *
CellShapeOut(i, fparam)
SaifEnum i;
MI_FPARAM   *fparam;
{ return(SaifEnumOut("CellShape", CellShape_enum, CellShapeN, i, fparam)); }

char *GridOrder_enum[] =                  /* <<<<<<<<<< GridOrder >>>>>>>>>> */
{ "rowOrder", "columnOrder", "boustrophedonByRow", "boustrophedonByColumn", };
#define GridOrderN (sizeof (GridOrder_enum) / sizeof (char *))

SaifEnum 
GridOrderIn(s, fparam)
mi_text   *s;
MI_FPARAM *fparam;
{ return(SaifEnumIn("GridOrder", GridOrder_enum, GridOrderN, s, fparam)); }

mi_text *
GridOrderOut(i, fparam)
SaifEnum i;
MI_FPARAM   *fparam;
{ return(SaifEnumOut("GridOrder", GridOrder_enum, GridOrderN, i, fparam)); }

char *IndexScheme_enum[] =              /* <<<<<<<<<< IndexScheme >>>>>>>>>> */
{ "sequentialArray", "interleavedArray", "runLengthEncoded", "nTree", };
#define IndexSchemeN (sizeof (IndexScheme_enum) / sizeof (char *))

SaifEnum
IndexSchemeIn(s, fparam)
mi_text   *s;
MI_FPARAM *fparam;
{ return(SaifEnumIn("IndexScheme", IndexScheme_enum, IndexSchemeN, s, fparam)); }

mi_text *
IndexSchemeOut(i, fparam)
SaifEnum i;
MI_FPARAM   *fparam;
{ return(SaifEnumOut("IndexScheme", IndexScheme_enum, IndexSchemeN,i,fparam)); }

char *LocationalQualifier_enum[] = /* <<<<<<<<<< LocationalQualifier >>>>>>>> */
{ "definite", "indefinite", "positionApproximate", "virtual", };
#define LocationalQualifierN (sizeof(LocationalQualifier_enum)/sizeof (char *))

SaifEnum
LocationalQualifierIn(s, fparam)
mi_text   *s;
MI_FPARAM *fparam;
{ return(SaifEnumIn("LocationalQualifier", LocationalQualifier_enum, LocationalQualifierN, s, fparam)); }

mi_text *
LocationalQualifierOut(i, fparam)
SaifEnum i;
MI_FPARAM   *fparam;
{ return(SaifEnumOut("LocationalQualifier", LocationalQualifier_enum, LocationalQualifierN,i,fparam)); }

char *OffsetStructure_enum[] =       /* <<<<<<<<<< OffsetStructure >>>>>>>> */
{ "none", "in", "out", };
#define OffsetStructureN (sizeof(OffsetStructure_enum)/sizeof (char *))

SaifEnum
OffsetStructureIn(s, fparam)
mi_text   *s;
MI_FPARAM *fparam;
{ return(SaifEnumIn("OffsetStructure", OffsetStructure_enum, OffsetStructureN, s, fparam)); }

mi_text *
OffsetStructureOut(i, fparam)
SaifEnum i;
MI_FPARAM   *fparam;
{ return(SaifEnumOut("OffsetStructure", OffsetStructure_enum, OffsetStructureN,i,fparam)); }

char *OriginReference_enum[] =      /* <<<<<<<<<< OriginReference >>>>>>>>>> */
{
     "centre",
     "nw",            "sw",              "ne",            "se",
     "nwTop",         "swTop",           "neTop",         "seTop",
     "nwBottom",      "swBottom",        "neBottom",      "seBottom",
     "nwFirst",       "swFirst",         "neFirst",       "seFirst",
     "nwLast",        "swLast",          "neLast",        "seLast",
     "nwTopFirst",    "swTopFirst",      "neTopFirst",    "seTopFirst",
     "nwBottomFirst", "swBottomFirst",   "neBottomFirst", "seBottomFirst",
     "nwTopLast",     "swTopLast",       "neTopLast",     "seTopLast",
     "nwBottomLast",  "swBottomLast",    "neBottomLast",  "seBottomLast",
};
#define OriginReferenceN (sizeof (OriginReference_enum) / sizeof (char *))

SaifEnum 
OriginReferenceIn(s, fparam)
mi_text   *s;
MI_FPARAM *fparam;
{ return(SaifEnumIn("OriginReference", OriginReference_enum, OriginReferenceN, s, fparam)); }

mi_text *
OriginReferenceOut(i, fparam)
SaifEnum i;
MI_FPARAM   *fparam;
{ return(SaifEnumOut("OriginReference", OriginReference_enum, OriginReferenceN, i, fparam)); }

char *PrimitiveInstanceType_enum[] = /* <<<<<<<< PrimitiveInstanceType >>>>> */
{ "string",    "octetString",       "bitString", "boolean", 
  "integer8",  "integer8Unsigned",  "integer16", "integer16Unsigned", 
  "integer32", "integer32Unsigned", "integer64", "integer64Unsigned",
  "real32",    "real64",            "real80", };
#define PrimitiveInstanceTypeN (sizeof(PrimitiveInstanceType_enum)/sizeof (char *))

SaifEnum
PrimitiveInstanceTypeIn(s, fparam)
mi_text   *s;
MI_FPARAM *fparam;
{ return(SaifEnumIn("PrimitiveInstanceType", PrimitiveInstanceType_enum, PrimitiveInstanceTypeN, s, fparam)); }

mi_text *
PrimitiveInstanceTypeOut(i, fparam)
SaifEnum i;
MI_FPARAM   *fparam;
{ return(SaifEnumOut("PrimitiveInstanceType", PrimitiveInstanceType_enum, PrimitiveInstanceTypeN,i,fparam)); }

char *ProjectionCategory_enum[] =       /* <<<<<<<< ProjectionCategory >>>>> */
{ "geodetic", "geographic", "other", };
#define ProjectionCategoryN (sizeof(ProjectionCategory_enum)/sizeof (char *))

SaifEnum
ProjectionCategoryIn(s, fparam)
mi_text   *s;
MI_FPARAM *fparam;
{ return(SaifEnumIn("ProjectionCategory", ProjectionCategory_enum, ProjectionCategoryN, s, fparam)); }

mi_text *
ProjectionCategoryOut(i, fparam)
SaifEnum i;
MI_FPARAM   *fparam;
{ return(SaifEnumOut("ProjectionCategory", ProjectionCategory_enum, ProjectionCategoryN,i,fparam)); }

/* Different satellites have different names for their instruments;
 * e.g., avhrr calls them "channels", landsat calls them "bands".
 * This list will have to change and grow. If the list is highly
 * dynamic, then this should not be an enumeration.
 */
char *SatelliteBands_enum[] =             /* <<<<<<  SatelliteBands  >>>>>> */
{ "band1", "band2", "band3", "band4", "band5", "band6",
  "chan1", "chan2", "chan3", "chan4", "chan5",
  "ndvi", "satellite_zenith", "solar_zenith", "relative_azimuth", 
  "image_source"};
#define SatelliteBandsN (sizeof (SatelliteBands_enum) / sizeof (char *))

SaifEnum
SatelliteBandsIn(s, fparam)
mi_text   *s;
MI_FPARAM *fparam;
{ return(SaifEnumIn("SatelliteBands", SatelliteBands_enum, SatelliteBandsN, s, fparam)); }

mi_text *
SatelliteBandsOut(i, fparam)
SaifEnum i;
MI_FPARAM   *fparam;
{ return(SaifEnumOut("SatelliteBands", SatelliteBands_enum, SatelliteBandsN, i, fparam)); }

char *UnitMeaning_enum[] =              /* <<<<<<<<<< UnitMeaning >>>>>>>>>> */
{
     "kilometres",  "metres",     "centimetres",   "millimetres",
     "micrometres", "nanometres", "nauticalMiles", "miles",
     "fathoms",     "feet",       "inches",        "degrees",
     "radians",     "years",      "months",        "weeks",
     "days",        "hours",      "minutes",       "seconds",
     "nanoseconds", "DigitalNumber",
     "fahrenheit",  "kelvin",     "centigrade",    "celsius",
     "other"
};
#define UnitMeaningN (sizeof (UnitMeaning_enum) / sizeof (char *))

SaifEnum 
UnitMeaningIn(s, fparam)
mi_text   *s;
MI_FPARAM *fparam;
{ return(SaifEnumIn("UnitMeaning", UnitMeaning_enum, UnitMeaningN,s,fparam)); }

mi_text *
UnitMeaningOut(i, fparam)
SaifEnum i;
MI_FPARAM   *fparam;
{ return(SaifEnumOut("UnitMeaning", UnitMeaning_enum,UnitMeaningN,i,fparam)); }

/* <<<<<<<<<<<<<<<<<<<<<<<<< Main SAIF Functions >>>>>>>>>>>>>>>>>>>>>>>>>>> */

typedef struct enum_list    /* for help functions */
{
     char enum_name[30];
     char **enum_list;
     int  enum_N;
} ENUM_LIST;

ENUM_LIST Enum_list[] =
{   "CellShape",            CellShape_enum,            CellShapeN,
    "GridOrder",            GridOrder_enum,            GridOrderN,
    "IndexScheme",          IndexScheme_enum,          IndexSchemeN,
    "LocationalQualifier",  LocationalQualifier_enum,  LocationalQualifierN,
    "OffsetStructure",      OffsetStructure_enum,      OffsetStructureN,
    "OriginReference",      OriginReference_enum,      OriginReferenceN,
    "PrimitiveInstanceType",PrimitiveInstanceType_enum,PrimitiveInstanceTypeN,
    "ProjectionCategory",   ProjectionCategory_enum,   ProjectionCategoryN,
    "SatelliteBands",       SatelliteBands_enum,       SatelliteBandsN,
    "UnitMeaning",          UnitMeaning_enum,          UnitMeaningN,
};
#define Enum_listN (sizeof (Enum_list) / sizeof (ENUM_LIST))

SaifEnum 
SaifEnumIn(ename, elist, num, s, fparam)
char      *ename;
char      **elist;
int       num;
mi_text   *s;
MI_FPARAM *fparam;
{
     char           buf[MAX_ENUM_SIZE], msg_buf[80], *r = "SaifEnumIn";
     int            i;
     SaifEnum       result = (SaifEnum) BAD_ENUM;
     MI_CONNECTION  *conn;

     if(MI_FP_ARGISNULL(fparam, 3)) {
          MI_FP_SETRETURNISNULL(fparam, MI_TRUE);
          return(NULL);
     }

     if ((conn = mi_open(NULL, NULL, NULL)) == (MI_CONNECTION *)NULL) {
          sprintf(msg_buf, "%s: could not get %s connection", r, DB_NAME);
          mi_db_error_raise(conn, MI_EXCEPTION, msg_buf);
     }

     mi_text_to_buffer(s, buf); /* convert mi_text to null-terminated buf */

#ifdef DEBUG
     sprintf(msg_buf, "%s: input for '%s' is '%s'", r, ename, buf);
     mi_db_error_raise(conn, MI_MESSAGE, msg_buf);
#endif

     for (i=0; i < num; i++) {
          if (!strcmp(buf, elist[i])) {
               result = (SaifEnum) i;
#ifdef DEBUG
               sprintf(msg_buf, "%s = %d", buf, result);
                  mi_db_error_raise(conn, MI_MESSAGE, msg_buf);
#endif
          }
     }

     if( result == (SaifEnum) BAD_ENUM ) {
          sprintf(msg_buf, "%s: '%s' is not a valid %s.", r, buf, ename);
          mi_db_error_raise(conn, MI_EXCEPTION, msg_buf);
     }

     if (mi_close(conn) == MI_ERROR) {
          sprintf(msg_buf, "%s: could not close %s connection", r, DB_NAME);
          mi_db_error_raise(conn, MI_MESSAGE, msg_buf);
     }

     return(result);
}

mi_text *
SaifEnumOut(ename, elist, num, val, fparam)
char        *ename;
char        **elist;
int         num;
SaifEnum    val;
MI_FPARAM   *fparam;
{
     char           msg_buf[100], *r = "SaifEnumOut";
     mi_text        *retval = NULL;
     MI_CONNECTION  *conn;

     if(MI_FP_ARGISNULL(fparam, 3)) {
          MI_FP_SETRETURNISNULL(fparam, MI_TRUE);
          return(NULL);
     }

     if ((conn = mi_open(NULL, NULL, NULL)) == (MI_CONNECTION *)NULL) {
          sprintf(msg_buf, "%s: could not get %s connection", r, DB_NAME);
          mi_db_error_raise(conn, MI_EXCEPTION, msg_buf);
     }

#ifdef DEBUG
     sprintf(msg_buf, "%s: %s = %d (%s)", r, ename, val, elist[val]);
     mi_db_error_raise(conn, MI_MESSAGE, msg_buf);
#endif

     if ( (val >= 0) && (val < num) ) 
          retval = mi_string_to_text(elist[val]);
     else { /* something is seriously screwed up if this happens */
          sprintf(msg_buf,"%s: '%d' out of range for %s.", r,val,ename);
          mi_db_error_raise(conn, MI_EXCEPTION, msg_buf);
     }

     if (mi_close(conn) == MI_ERROR) {
          sprintf(msg_buf, "%s: could not close %s connection", r, DB_NAME);
          mi_db_error_raise(conn, MI_MESSAGE, msg_buf);
     }

     return (retval);
}

mi_text *
SaifEnumHelp_l()
{
     mi_text    *retval;
     int        data_size, i;

     /* Allocation size = each column size plus a space separator.
      * Leave the extra space for the last column so we can use sprintf
      * and avoid whomping illustrious memory.
      */
     for (data_size = 0, i = 0; i < Enum_listN; i++)
          data_size += strlen(Enum_list[i].enum_name) + 1;

     retval = (mi_text *) mi_alloc(MI_VARALLOC(data_size));
     MI_SET_VARLEN(retval, data_size);

     for(i = 0; i < Enum_listN; i++) {
          if(i == 0)
               sprintf(MI_GET_VARDATA(retval), "%s", Enum_list[i].enum_name);
          else
               sprintf(MI_GET_VARDATA(retval), "%s %s", 
                    MI_GET_VARDATA(retval), Enum_list[i].enum_name);
     }

     return (retval);
}

mi_text *
SaifEnumHelp(ename, fparam)
mi_text   *ename;
MI_FPARAM *fparam;
{
     mi_text *retval;
     char    *r = "SaifEnumHelp",
             *str,                  /* pointer to the enumeration name */
             **elist = NULL,        /* pointer to the enum's list of values */
             msg_buf[100];
     int     num=0, data_size, esize, i;

     MI_CONNECTION  *conn;

     if(MI_FP_ARGISNULL(fparam, 0)) {
          MI_FP_SETRETURNISNULL(fparam, MI_TRUE);
          return(NULL);
     }

     if ((conn = mi_open(NULL, NULL, NULL)) == (MI_CONNECTION *)NULL) {
          sprintf(msg_buf, "%s: could not get %s connection", r, DB_NAME);
          mi_db_error_raise(conn, MI_EXCEPTION, msg_buf);
     }

     str = mi_text_to_string(ename);
     esize = MI_GET_TEXTLEN(ename);

     for(i=0; i < Enum_listN; i++) {
          if(!strncmp(str, Enum_list[i].enum_name, esize)) {
               num = Enum_list[i].enum_N;
               elist = Enum_list[i].enum_list;
          }
     }

     if(num == 0) {
          sprintf(msg_buf,
               "%s: '%s' is not an implemented SAIF enumeration.", 
               r, str);
          mi_db_error_raise(conn, MI_EXCEPTION, msg_buf);
     }

     /* Allocation size = each column size plus a space separator.
      * Leave the extra space at the end so we can use sprintf.
      */
     for (data_size = 0, i = 0; i < num; i++)
          data_size += strlen(elist[i]) + 1;

     retval = (mi_text *) mi_alloc(MI_VARALLOC(data_size));
     MI_SET_VARLEN(retval, data_size);

     for(i = 0; i < num; i++) {
          if(i == 0)
               sprintf(MI_GET_VARDATA(retval), "%s", elist[i]);
          else
               sprintf(MI_GET_VARDATA(retval), "%s %s", 
                    MI_GET_VARDATA(retval), elist[i]);
     }

     if (mi_close(conn) == MI_ERROR) {
          sprintf(msg_buf, "%s: could not close %s connection", r, DB_NAME);
          mi_db_error_raise(conn, MI_MESSAGE, msg_buf);
     }

     return (retval);
}

/*============ integer ops for implementing enums as base types ============ */
mi_boolean
SaifEnumEQ(a1, a2, fparam)
SaifEnum   a1, a2;
MI_FPARAM  *fparam;
{
     if( MI_FP_ARGISNULL(fparam, 0) || MI_FP_ARGISNULL(fparam, 1) ) {
          MI_FP_SETRETURNISNULL(fparam, MI_TRUE);
          return(NULL);
     }
     return((mi_boolean) a1 == a2);
}

mi_boolean
SaifEnumNE(a1, a2, fparam)
SaifEnum   a1, a2;
MI_FPARAM  *fparam;
{
     if( MI_FP_ARGISNULL(fparam, 0) || MI_FP_ARGISNULL(fparam, 1) ) {
          MI_FP_SETRETURNISNULL(fparam, MI_TRUE);
          return(NULL);
     }
     return((mi_boolean) a1 != a2);
}

mi_boolean
SaifEnumLT(a1, a2, fparam)
SaifEnum   a1, a2;
MI_FPARAM  *fparam;
{
     if( MI_FP_ARGISNULL(fparam, 0) || MI_FP_ARGISNULL(fparam, 1) ) {
          MI_FP_SETRETURNISNULL(fparam, MI_TRUE);
          return(NULL);
     }
     return((mi_boolean) a1 < a2);
}

mi_boolean
SaifEnumLE(a1, a2, fparam)
SaifEnum   a1, a2;
MI_FPARAM  *fparam;
{
     if( MI_FP_ARGISNULL(fparam, 0) || MI_FP_ARGISNULL(fparam, 1) ) {
          MI_FP_SETRETURNISNULL(fparam, MI_TRUE);
          return(NULL);
     }
     return((mi_boolean) a1 <= a2);
}

mi_boolean
SaifEnumGT(a1, a2, fparam)
SaifEnum   a1, a2;
MI_FPARAM  *fparam;
{
     if( MI_FP_ARGISNULL(fparam, 0) || MI_FP_ARGISNULL(fparam, 1) ) {
          MI_FP_SETRETURNISNULL(fparam, MI_TRUE);
          return(NULL);
     }
     return((mi_boolean) a1 > a2);
}

mi_boolean
SaifEnumGE(a1, a2, fparam)
SaifEnum   a1, a2;
MI_FPARAM  *fparam;
{
     if( MI_FP_ARGISNULL(fparam, 0) || MI_FP_ARGISNULL(fparam, 1) ) {
          MI_FP_SETRETURNISNULL(fparam, MI_TRUE);
          return(NULL);
     }
     return((mi_boolean) a1 >= a2);
}
